"""
fridex/connection/communication/_test_communication.py

Project: Fridrich-Connection
Created: 13.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

import unittest

from ._client_connection import ClientConnection
from ._client_handler import ClientHandler


##################################################
#                     Code                       #
##################################################

class CommunicationTest(unittest.TestCase):
    """
    Basic client - server communication test
    """
    __client: ClientConnection
    __server: ClientHandler

    def setUp(self) -> None:
        """
        Setup client and server
        """

    def tearDown(self) -> None:
        """
        Delete client and server
        """



"""
deamons/connection/protocol/_subscription.py

Project: Fridrich-Connection
Created: 26.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import Callable, Any, TypedDict, Literal, Type
from concurrent.futures import ThreadPoolExecutor


from ._types import BulkDict, DATAUNIT
from ._protocol import Protocol
from ._cache import Cache


##################################################
#                     Code                       #
##################################################

class SubscriptionRequest(TypedDict):
    action: Literal["add", "delete"]
    value: Type[DATAUNIT] | int


CALLBACK_TYPE = Callable[[DATAUNIT], Any]


class SubscriptionSave(TypedDict):
    callback: CALLBACK_TYPE
    req_dict: Type[DATAUNIT]


class SubscriptionProtocol:
    """
    Protocol for subscriptions
    """
    __protocol: Protocol
    __cache: Cache | None

    __thread_pool: ThreadPoolExecutor
    __subscriptions: dict[int, SubscriptionSave]

    ADD_RELATED_SUB_CALLBACK_TYPE = Callable[[int, DATAUNIT], Any]
    DELETE_RELATED_SUB_CALLBACK_TYPE = Callable[[int], Any]
    SEND_SUB_CALLBACK_TYPE = Callable[[str], Any]

    __add_related_sub_callback: ADD_RELATED_SUB_CALLBACK_TYPE | None
    __delete_related_sub_callback: DELETE_RELATED_SUB_CALLBACK_TYPE | None
    __send_sub_callback: SEND_SUB_CALLBACK_TYPE | None

    def __init__(
            self,
            id_range: range,
            thread_pool: ThreadPoolExecutor,
            add_related_sub_callback: ADD_RELATED_SUB_CALLBACK_TYPE | None,
            delete_related_sub_callback: DELETE_RELATED_SUB_CALLBACK_TYPE | None,
            send_sub_callback: SEND_SUB_CALLBACK_TYPE | None,
            cache: Cache | None = None
    ) -> None:
        """
        Create subscription protocol
        :param id_range: ID range to use for this subprotocol
        :param thread_pool: ThreadPool to execute callbacks
        :param add_related_sub_callback: Callback when an add subscription request comes in
        :param delete_related_sub_callback: Callback when a delete subscription request comes in
        :param send_sub_callback: Callback to send subscription data
        :param cache: Optional cache
        """
        self.__protocol = Protocol("sub", id_range)

        self.__thread_pool = thread_pool
        self.__add_related_sub_callback = add_related_sub_callback
        self.__delete_related_sub_callback = delete_related_sub_callback
        self.__send_sub_callback = send_sub_callback
        self.__cache = cache

    def __response(
            self,
            data: DATAUNIT,
            id_: int
    ) -> str:
        """
        General response encapsulation
        :param data: Response dictonary
        :param id_: ID of the conversation
        :return: Raw string to send
        """
        self.__protocol.response_add(data, id_)
        return self.__protocol.response_get()

    def __request(self, data: SubscriptionRequest) -> str:
        """
        General request encapsulation
        :param data: Subscription request
        :return: Request string to send
        """
        self.__protocol.request_add(data)
        return self.__protocol.request_get()

    def add_subscription(
            self,
            callback: CALLBACK_TYPE,
            request_dict: DATAUNIT
    ) -> tuple[int, str]:
        """
        Add a new subscription
        :param callback: Callback when value is updated
        :param request_dict: Same dictonary as a normal request to use
        :return: Subscription ID and String to send
        """
        message: str = self.__request({"action": "add", "value": request_dict})
        self.__subscriptions[self.__protocol.id_count] = {"callback": callback, "req_dict": request_dict}

        return self.__protocol.id_count, message

    def remove_subscription(
            self,
            subscription_id: int
    ) -> str:
        """
        Remove a subscription
        :param subscription_id: ID of the subscription
        """
        self.__subscriptions.pop(subscription_id)
        return self.__request({"action": "delete", "value": subscription_id})

    def _response_subscription(
            self,
            id_: int,
            value: dict[str | int | float | bool | None, any]
    ) -> str:
        """
        Respone to a subscription (update value)
        :param id_: ID of the subscription
        :param value: New value
        :return: Response string to send
        """
        return self.__response(value, id_)

    def provide_data(self, req_dict: DATAUNIT, value: DATAUNIT) -> None:
        """
        Check if data is needed in a subscription
        :param req_dict: Same dictonary as a normal request to use
        :param value: New value
        """
        for sub in self.__subscriptions:
            if self.__subscriptions[sub]["req_dict"] == req_dict:
                self.__send_sub_callback(self._response_subscription(sub, value))

    def process_response(self, message: BulkDict) -> None:
        """
        These responses are coming without a request
        :param message: Response message
        """
        submessage = message["data"][0]
        if self.__cache:
            self.__cache.set(self.__subscriptions[submessage["id"]]["req_dict"], submessage["data"])
        self.__thread_pool.submit(self.__subscriptions[submessage["id"]]["callback"], submessage["data"])

    def process_request(self, message: BulkDict) -> None:
        """
        Process requests for new subscription management
        Subscription requests do not get an immidiate response
        :param message: Request message
        """
        submessage = message["data"][0]
        match submessage["data"]["action"]:
            case "add":
                if self.__add_related_sub_callback is not None:
                    self.__add_related_sub_callback(submessage["id"], submessage["data"]["value"])
            case "delete":
                if self.__delete_related_sub_callback is not None:
                    self.__delete_related_sub_callback(submessage["data"]["value"])



"""
deamons/connection/protocol/__init__.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""
from ._types import MessageDict, BulkDict, KINDS, DIRECTIONS, DATAUNIT
from ._communication import CommunicationProtocol, CommunicationData
from ._subscription import SubscriptionProtocol, SubscriptionRequest
from ._protocol import Protocol, MessageToLongError
from ._protocol_interface import ProtocolInterface
from ._control import ControlData, ControlProtocol
from ._cache import CacheEntry, Cache
from ._data import DataProtocol



"""
fridex/connection/communication/_client_handler.py

Project: Fridrich-Connection
Created: 12.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from concurrent.futures import ThreadPoolExecutor
from time import sleep
import socket

from ..protocol import ProtocolInterface, DATAUNIT
from ._server_connection import ServerConnection


##################################################
#                     Code                       #
##################################################

class ClientHandler(socket.socket):
    """
    Serverside ClientHandler
    """
    __clients: list[ServerConnection]

    __request_callback: ProtocolInterface.REQUEST_CALLBACK_TYPE
    __rework_callback: ProtocolInterface.REWORK_CALLBACK_TYPE
    __add_sub_callback: ProtocolInterface.ADD_RELATED_SUB_CALLBACK_TYPE
    __del_sub_callback: ProtocolInterface.DELETE_RELATED_SUB_CALLBACK_TYPE

    def __init__(
            self,
            request_callback: ProtocolInterface.REQUEST_CALLBACK_TYPE,
            rework_callback: ProtocolInterface.REWORK_CALLBACK_TYPE,
            add_sub_callback: ProtocolInterface.ADD_RELATED_SUB_CALLBACK_TYPE,
            del_sub_callback: ProtocolInterface.DELETE_RELATED_SUB_CALLBACK_TYPE,
            port: int | None = 4205,
    ) -> None:
        """
        Create server with client accept handler
        :param request_callback: Callback to get information for data requests
        :param add_sub_callback: Callback when an add subscription request comes in
        :param del_sub_callback: Callback when a delete subscription request comes in
        :param port: Port to open the server on
        """
        super().__init__(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(("0.0.0.0", port))
        self.listen()

        self.__request_callback = request_callback
        self.__rework_callback = rework_callback
        self.__add_sub_callback = add_sub_callback
        self.__del_sub_callback = del_sub_callback

        self.__clients = []
        self.__threadpool = ThreadPoolExecutor(max_workers=1)

        self.__threadpool.submit(self.__accept_clients)

    __threadpool: ThreadPoolExecutor

    def __accept_clients(self) -> None:
        """
        Loop accept client connections
        """
        while True:
            sock, address = self.accept()

            self.__clients.append(ServerConnection(sock,
                                                   request_callback=self.__request_callback,
                                                   rework_callback=self.__rework_callback,
                                                   add_sub_callback=self.__add_sub_callback,
                                                   del_sub_callback=self.__del_sub_callback))
            sleep(0.1)

    def provide_data(self, req_dict: DATAUNIT, value: DATAUNIT) -> None:
        """
        Provide data for subscription to all clients
        :param req_dict: Same dictonary as a normal request to use
        :param value: New value
        """
        for client in self.__clients:
            client.provide_data(req_dict, value)




"""
deamons/connection/communication/__init__.py

Project: Fridrich-Connection
Created: 26.05.2023
Author: Lukas Krahbichler
"""

from ._client_connection import ClientConnection
from ._server_connection import ServerConnection
from ._client_handler import ClientHandler



"""
deamons/connection/communication/_base_connection.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Any, Literal
from datetime import datetime, timedelta
from time import sleep
import socket
import select

from ..encryption import CryptionService, CryptionMethod, CRYPTION_METHODS
from ..protocol import BulkDict, ProtocolInterface, Protocol
from ..protocol import CommunicationProtocol


##################################################
#                     Code                       #
##################################################

class BaseConnection:
    """
    This represents a connection between server and client
    """
    __socket: socket.socket
    __timeout: int
    __start_time: datetime
    __lease_time: datetime

    _thread_pool: ThreadPoolExecutor
    _protocol: ProtocolInterface

    _cryption: CryptionMethod
    _new_cryption: CryptionMethod | None

    _send_data: list[str]
    _send_communication: list[Literal["key"] | CRYPTION_METHODS]
    _respond_communication: str | None

    __STATES = Literal["open", "paused", "prewait", "waiting", "afterwait", "closed"]
    __state: __STATES | Literal["all"]
    __state_callbacks: dict[int, tuple[__STATES, Callable[[], Any]]]

    def __init__(
            self,
            conn: socket.socket,
            request_callback: ProtocolInterface.REQUEST_CALLBACK_TYPE,
            rework_callback: ProtocolInterface.REWORK_CALLBACK_TYPE,
            add_sub_callback: ProtocolInterface.ADD_RELATED_SUB_CALLBACK_TYPE | None = None,
            del_sub_callback: ProtocolInterface.DELETE_RELATED_SUB_CALLBACK_TYPE | None = None,
            timeout: int = 10,
            packet_size: int = 1,
    ) -> None:
        """
        Create connection
        :param conn: Socket
        :param request_callback: Callback to get information for data requests
        :param rework_callback: Callback to rework result before setting to future
        :param add_sub_callback: Callback when an add subscription request comes in
        :param del_sub_callback: Callback when a delete subscription request comes in
        :param timeout: Connection leasetime if no response on ping
        """
        if timeout < 2:
            timeout = 2
        self.__timeout = timeout

        self.__packet_size = packet_size

        self.__socket = conn
        self.__socket.settimeout(0.1)

        self.__start_time = datetime.now()
        self.__lease_time = self.__start_time + timedelta(seconds=self.__timeout)

        self.__state = "open"
        self.__state_callbacks = {}

        self._send_data = []
        self._send_communication = []
        self._respond_communication = None

        self._thread_pool = ThreadPoolExecutor(max_workers=2)
        self._cryption = CryptionService.new_cryption()
        self._new_cryption = None

        self._protocol = ProtocolInterface(
            data_callback=request_callback,
            rework_callback=rework_callback,
            new_key_callback=self._cryption.new_key,
            set_key_callback=self._cryption.set_key,
            control_callback=self.__confirm_control,
            ping_callback=self.__ping_confirm,
            max_bytes_callback=Protocol.set_max_bytes,
            cryption_req_callback=self.__cryption_request,
            cryption_res_callback=self.__cryption_confirm,
            pause_connection_callback=lambda: self._set_state("paused"),
            resume_connection_callback=lambda: self._set_state("open"),
            thread_pool=self._thread_pool,
            add_related_sub_callback=add_sub_callback,
            delete_related_sub_callback=del_sub_callback,
            send_sub_callback=self.send
        )

        self._thread_pool.submit(self.__loop)

    def __confirm_control(self) -> None:
        """
        Confirm control message that requires pausing the connection
        """
        if self.__state == "afterwait":
            self._set_state("open")
        else:
            self._set_state("prewait")

    def __cryption_request(
            self,
            encryption: CRYPTION_METHODS,
            key: str
    ) -> tuple[CommunicationProtocol.NEW_KEY_CALLBACK_TYPE, CommunicationProtocol.SET_KEY_CALLBACK_TYPE]:
        """
        Callback when cryption exchange is requested
        :param encryption: New encryption type string
        :param key: New key
        :return: New cryption callbacks for ControlProtocol
        """
        self._cryption = CryptionService.new_cryption(encryption)
        self._cryption.set_key(key)
        return self._cryption.new_key, self._cryption.set_key

    def __cryption_confirm(self, key: str) -> tuple[CommunicationProtocol.NEW_KEY_CALLBACK_TYPE, CommunicationProtocol.SET_KEY_CALLBACK_TYPE]:
        """
        Callback when cryption exchange was successful
        :param key: New key
        :return: New cryption callbacks for ControlProtocol
        """
        self._cryption.set_key(key)

        return self._cryption.new_key, self._cryption.set_key

    def __ping_confirm(self) -> None:
        """
        Callback when ping request gets a response
        """
        self.__lease_time = datetime.now() + timedelta(seconds=self.__timeout)

    def __loop(self) -> None:
        next_alive: datetime = datetime.now()

        while True:
            if self.__state == "open":
                # Request ping
                if datetime.now() > next_alive:
                    self._send_data.append(self._protocol.control.request_alive())
                    next_alive = datetime.now() + timedelta(seconds=self.__timeout)

                # Close connection if leased
                if datetime.now() > self.__lease_time + timedelta(seconds=2):
                    self.__socket.close()
                    return

            # Sending
            to_send: list[str] = []

            # Pause other side
            if self.__state == "open" and self._send_communication:
                self._set_state("waiting")
                to_send = [self._protocol.communication.request_pause()]

            if self.__state == "prewait" and not self._send_communication:
                self._set_state("afterwait")
                to_send = [self._protocol.communication.request_resume()]

            # Choose what to send
            match self.__state:
                case "prewait":
                    send = self._send_communication.pop(0)
                    match send:
                        case "key":
                            to_send = [self._protocol.communication.request_key_exchange()]

                        case "private_public" | "fernet":
                            self._new_cryption = CryptionService.new_cryption(send)
                            to_send = [
                                self._protocol.communication.request_crpytion(
                                    send,
                                    self._new_cryption.new_key()
                                )
                            ]
                    self._set_state("waiting")

                case "waiting" | "afterwait":
                    ...

                case "paused" | "open":
                    if self._respond_communication:
                        to_send.append(self._respond_communication)
                        self._respond_communication = None

                    if self.__state == "open":
                        to_send += self._send_data.copy()
                        self._send_data = []

            # Sending
            for send in to_send:
                size_send: bytes = len(send).to_bytes(length=Protocol.max_bytes, byteorder="big")
                self.__socket.send(self._cryption.encrypt(size_send + send.encode("UTF-8")))

            if self._new_cryption:
                self._cryption = self._new_cryption
                self._new_cryption = None

            # Receiving
            encrypted_buffer: bytes = bytes()
            while True:
                ready_to_read, _, _ = select.select([self.__socket], [], [], 0)
                if ready_to_read:
                    recv: bytes = self.__socket.recv(self.__packet_size)
                    if not recv:
                        break
                    encrypted_buffer += recv
                else:
                    break

            # Work out inividual messages
            message_buffer: bytes = self._cryption.decrypt(encrypted_buffer)
            recv_messages: list[BulkDict] = []

            while message_buffer != b'':
                size_recv: int = int.from_bytes(bytes=message_buffer[:Protocol.max_bytes], byteorder="big")

                message_bytes = message_buffer[Protocol.max_bytes:Protocol.max_bytes + size_recv]
                recv_messages.append(self._protocol.decapsulate(message_bytes))

                message_buffer = message_buffer[Protocol.max_bytes + size_recv:]

            # Process received messages
            for message in recv_messages:
                match message["direction"]:
                    case "request":
                        match message["kind"]:
                            case "data":
                                self._send_data.append(self._protocol.data.process_request(message))
                            case "sub":
                                self._protocol.subscription.process_request(message)
                            case "com":
                                self._respond_communication = self._protocol.communication.process_request(message)
                            case "con":
                                con_res = self._protocol.control.process_request(message)
                                if con_res:
                                    self._send_data.append(con_res)

                    case "response":
                        match message["kind"]:
                            case "data":
                                self._protocol.data.process_response(message)
                            case "sub":
                                self._protocol.subscription.process_response(message)
                            case "con":
                                self._protocol.control.process_response(message)
                            case "com":
                                self._protocol.communication.process_response(message)

            sleep(0.05)

    def close(self) -> None:
        """
        Close connection
        """
        self._set_state("closed")
        self._thread_pool.shutdown(wait=False, cancel_futures=True)
        self.__socket.close()

    def send(self, message: str) -> None:
        """
        Send a message
        :param message: Raw string message
        """
        if self.__state == "open":
            self._send_data.append(message)
            return

        raise ConnectionError("Connection is not in state 'open'.")

    def send_key_exchange(self) -> None:
        """
        Send key exchange message
        """
        self._send_communication.append("key")

    def send_cryption_change(self, cryption: CRYPTION_METHODS) -> None:
        """
        Send cryption change message
        """
        self._send_communication.append(cryption)

    @property
    def state(self) -> Literal["init", "open", "closed"]:
        """
        :return: Current connection state
        """
        return self.__state

    def _set_state(self, value: __STATES) -> None:
        """
        Set current state and go through callbacks
        :param value: Value to set to
        """
        self.__state = value

        for cb_id, (state, callback) in self.__state_callbacks.items():
            if state == self.__state:
                callback()

        if self.__state == "open":
            self.__lease_time = datetime.now() + timedelta(seconds=self.__timeout+1)

    def join_state(self, state: __STATES, _time_delta: float | None = 0.1) -> None:
        """
        Wait until certain state is reached
        :param state: State to wait for
        :param _time_delta: Time interval for checking
        """
        while self.__state != state:
            sleep(_time_delta)

    def callback_state(self, state: __STATES | Literal["all"], callback: Callable[[], Any]) -> int:
        """
        Get a callback when certain state is reached
        :param state: On what state
        :param callback: Function to call
        :return: Callback ID
        """
        num: int = max(self.__state_callbacks.keys()) if self.__state_callbacks else 0
        self.__state_callbacks[num] = (state, callback)
        return num

    def remove_callback_state(self, callback_id: int) -> None:
        """
        Remove certain state callback
        :param callback_id: ID of the callback
        """
        self.__state_callbacks.pop(callback_id)

    def remove_all_callbacks(self) -> None:
        """
        Removes all current callbacks
        """
        self.__state_callbacks = {}



"""
src/Fridrich-Connection/__init__.py

Project: Fridrich-Connection
Created: 23.05.2023
Author: Lukas Krahbichler
"""

from .communication import *
from .encryption import *
from .protocol import *



"""
deamons/connection/protocol/_cache.py

Project: Fridrich-Connection
Created: 01.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import TypedDict, Type
from datetime import datetime, timedelta

from ._types import DATAUNIT


##################################################
#                     Code                       #
##################################################

class CacheEntry(TypedDict):
    value: Type[DATAUNIT]
    lease_time: datetime


class Cache:
    """
    Caching for subscriptions
    """
    __values: dict[DATAUNIT, CacheEntry]
    __lifetime: float

    def __init__(self, lifetime: float | None = 30.0) -> None:
        """
        Create empty cache
        :param lifetime: The time after a value is dropped if it's not used
        """
        self.__lifetime = lifetime

        self.__values = {}

    def _kill_dead(self) -> None:
        """
        Check all leasetimes and drop old entries
        """
        to_drop: list[DATAUNIT] = []

        for key, value in self.__values.items():
            if datetime.now() > value["lease_time"]:
                to_drop.append(key)

        for drop in to_drop:
            self.__values.pop(drop)

    def set(
        self,
        key: DATAUNIT,
        value: DATAUNIT
    ) -> None:
        """
        Update or create an entry
        :param key: Key to associate the value
        :param value: Data to save
        """
        self.__values[key] = {"value": value, "lease_time": datetime.now() + timedelta(seconds=self.__lifetime)}

        self._kill_dead()

    def get(
        self,
        key: DATAUNIT
    ) -> DATAUNIT | None:
        """
        Get the value of an entry
        :param key: The key to look for
        :return: Value if entry with key exists
        """
        if key in self.__values:
            self.__values[key]["lease_time"] = datetime.now() + timedelta(seconds=self.__lifetime)

            return self.__values[key]
        return None



"""
deamons/connection/encryption/_private_public.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.backends import default_backend

from ._cryption_method import CryptionMethod


##################################################
#                     Code                       #
##################################################

class PrivatePublicCryption(CryptionMethod):
    """
    A high security private-public-keypair encryption
    """

    __own_private_key: rsa.RSAPrivateKey | None
    __own_public_key: rsa.RSAPublicKey | None
    __foreign_public_key: rsa.RSAPublicKey | None
    __public_str: str | None

    __key_size: int

    def __init__(self) -> None:
        """
        Create public-private cryption model
        """
        self.__foreign_public_key = None
        self.__own_private_key = None
        self.__own_public_key = None
        self.__public_str = None
        self.__key_size = (256 + 66) * 8

    def encrypt(self, message: bytes) -> bytes:
        if self.__foreign_public_key is not None:
            size: int = (self.__foreign_public_key.key_size // 8) - 66
            splits: list[bytes] = [message[i:i + size] for i in range(0, len(message), size)]
            encrypted_result: bytes = b''

            for split in splits:
                encrypted_result += self.__foreign_public_key.encrypt(
                    split,
                    padding.OAEP(
                        mgf=padding.MGF1(algorithm=hashes.SHA256()),
                        algorithm=hashes.SHA256(),
                        label=None
                    )
                )

        else:
            encrypted_result = message

        return encrypted_result

    def decrypt(self, message: bytes) -> bytes:
        decrypted_result: bytes = b""

        if self.__own_public_key is not None:

            size: int = (self.__own_private_key.key_size // 8)
            splits: list[bytes] = [message[i:i+size] for i in range(0, len(message), size)]
            for split in splits:
                decrypted_result += self.__own_private_key.decrypt(
                    split,
                    padding.OAEP(
                        mgf=padding.MGF1(algorithm=hashes.SHA256()),
                        algorithm=hashes.SHA256(),
                        label=None
                    )
                )
        else:
            decrypted_result = message

        return decrypted_result

    def get_key(self) -> str:
        if not self.__public_str:
            self.new_key()
        return self.__public_str

    def set_key(self, key: str) -> None:
        self.__foreign_public_key = serialization.load_pem_public_key(
            data=key.encode("ASCII"),
            backend=default_backend()
        )

    def new_key(self) -> str:
        self.__own_private_key = rsa.generate_private_key(public_exponent=65537, key_size=self.__key_size)
        self.__own_public_key = self.__own_private_key.public_key()

        self.__public_str = self.__own_public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode("ASCII")

        return self.get_key()



"""
deamons/connection/protocol/_protocol_interface.py

Project: Fridrich-Connection
Created: 31.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from concurrent.futures import ThreadPoolExecutor
from json import loads

from ._subscription import SubscriptionProtocol
from ._communication import CommunicationProtocol
from ._control import ControlProtocol
from ._data import DataProtocol
from ._types import BulkDict
from ._cache import Cache


##################################################
#                     Code                       #
##################################################


class ProtocolInterface:
    """
    Interface to access all protocols
    """
    __cache: Cache
    __data: DataProtocol
    __control: ControlProtocol
    __communication: CommunicationProtocol
    __subscription: SubscriptionProtocol

    REQUEST_CALLBACK_TYPE = DataProtocol.REQUEST_CALLBACK_TYPE
    REWORK_CALLBACK_TYPE = DataProtocol.REWORK_CALLBACK_TYPE
    ADD_RELATED_SUB_CALLBACK_TYPE = SubscriptionProtocol.ADD_RELATED_SUB_CALLBACK_TYPE | None
    DELETE_RELATED_SUB_CALLBACK_TYPE = SubscriptionProtocol.DELETE_RELATED_SUB_CALLBACK_TYPE | None

    def __init__(
            self,
            data_callback: DataProtocol.REQUEST_CALLBACK_TYPE,
            rework_callback: DataProtocol.REWORK_CALLBACK_TYPE,
            new_key_callback: CommunicationProtocol.NEW_KEY_CALLBACK_TYPE,
            set_key_callback: CommunicationProtocol.SET_KEY_CALLBACK_TYPE,
            control_callback: CommunicationProtocol.CONTROL_CALLBACK,
            ping_callback: ControlProtocol.PING_CALLBACK_TYPE,
            max_bytes_callback: CommunicationProtocol.MAX_BYTES_CALLBACK_TYPE,
            cryption_req_callback: CommunicationProtocol.CRYPTION_REQ_CALLBACK_TYPE,
            cryption_res_callback: CommunicationProtocol.CRYPTION_RES_CALLBACK_TYPE,
            pause_connection_callback: CommunicationProtocol.PAUSE_CONNECTION_CALLBACK_TYPE,
            resume_connection_callback: CommunicationProtocol.RESUME_CONNECTION_CALLBACK_TYPE,
            thread_pool: ThreadPoolExecutor,
            add_related_sub_callback: SubscriptionProtocol.ADD_RELATED_SUB_CALLBACK_TYPE | None = None,
            delete_related_sub_callback: SubscriptionProtocol.DELETE_RELATED_SUB_CALLBACK_TYPE | None = None,
            send_sub_callback: SubscriptionProtocol.SEND_SUB_CALLBACK_TYPE | None = None,
            max_bytes: int = 4
    ) -> None:
        """
        Create all protocols
        :param data_callback: Callback to get information for requests
        :param rework_callback: Callback to rework data result before setting to future
        :param new_key_callback: Callback to generate a new key and get it
        :param set_key_callback: Callback when a new key is received
        :param control_callback: Callback when control response is received that requires pausing the connection
        :param ping_callback: Callback when ping response is received
        :param max_bytes_callback: Callback to set new max bytes
        :param cryption_req_callback: Callback to set new encryption with key.
        :param cryption_res_callback: Callback to confirm cryption change and set key
        :param pause_connection_callback: Callback to pause connection if requested
        :param resume_connection_callback: Callback to resume connection if requested
        :param thread_pool: ThreadPool to execute callbacks
        :param add_related_sub_callback: Callback when an add subscription request comes in
        :param delete_related_sub_callback: Callback when a delete subscription request comes in
        :param send_sub_callback: Callback to send subscription data
        :param max_bytes: Number of bytes to communicate length
        """
        DataProtocol.set_max_bytes(max_bytes)

        self.__cache = Cache()

        self.__data = DataProtocol(range(0, 999), request_callback=data_callback,
                                   rework_callback=rework_callback, cache=self.__cache)
        self.__control = ControlProtocol(range(1000, 1999), ping_callback)
        self.__communication = CommunicationProtocol(range(2000, 2999), new_key_callback, set_key_callback,
                                                     control_callback, max_bytes_callback,
                                                     cryption_req_callback, cryption_res_callback,
                                                     pause_connection_callback, resume_connection_callback)
        self.__subscription = SubscriptionProtocol(range(3000, 3999), thread_pool,
                                                   add_related_sub_callback, delete_related_sub_callback,
                                                   send_sub_callback, cache=self.__cache)

    def decapsulate(self, messages: bytes) -> BulkDict:  # noqa
        """
        Decode and convert to dict
        :param messages: Raw bytes without length header
        :return: Dictonary with data
        """
        return loads(messages.decode("UTF-8"))

    @property
    def data(self) -> DataProtocol:
        """
        :return: DataProtocol instance
        """
        return self.__data

    @property
    def control(self) -> ControlProtocol:
        """
        :return: ControlProtocol instance
        """
        return self.__control

    @property
    def communication(self) -> CommunicationProtocol:
        """
        :return: CommunicationProtocol instance
        """
        return self.__communication

    @property
    def subscription(self) -> SubscriptionProtocol:
        """
        :return: SubscriptionProtocol instance
        """
        return self.__subscription



"""
fridex/connection/encryption/_test_private_public.py

Project: Fridrich-Connection
Created: 12.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from ._supertest_cryption import CryptionTest


##################################################
#                     Code                       #
##################################################

class PrivatePublicTest(CryptionTest):
    """
    Test PrivatePublic encryption
    """
    _cryption = "private_public"

    def test_left_key(self) -> None:
        """
        Test cryption with only left key generated
        """
        self._right.set_key(self._left.get_key())
        self.dual_side_exchange()

        self.assertEqual(self._raw_left, self._encrypted_left)
        self.assertNotEqual(self._raw_right, self._encrypted_right)

    def test_right_key(self) -> None:
        """
        Test cryption with only __right generated
        """
        self._left.set_key(self._right.get_key())
        self.dual_side_exchange()

        self.assertNotEqual(self._raw_left, self._encrypted_left)
        self.assertEqual(self._raw_right, self._encrypted_right)



"""
fridex/connection/encryption/_fernet.py

Project: Fridrich-Connection
Created: 12.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from cryptography.fernet import Fernet

from ._cryption_method import CryptionMethod


##################################################
#                     Code                       #
##################################################

class FernetCryption(CryptionMethod):
    """
    Simple fernet encryption
    """
    __own_cryption: Fernet | None
    __foreign_cryption: Fernet | None

    __own_key: bytes | None
    __foreign_key: bytes | None

    def __init__(self) -> None:
        """
        Create FernetCryption
        """
        self.__own_cryption = None
        self.__foreign_cryption = None
        self.__own_key = None
        self.__foreign_key = None

    def encrypt(self, message: bytes) -> bytes:
        if self.__foreign_cryption:
            return self.__foreign_cryption.encrypt(message)
        return message

    def decrypt(self, message: bytes) -> bytes:
        if self.__own_cryption and message != b'':
            return self.__own_cryption.decrypt(message)

        return message

    def get_key(self) -> str:
        if not self.__own_key:
            self.new_key()
        return self.__own_key.decode("UTF-8")

    def set_key(self, key: str) -> None:
        self.__foreign_key = key.encode("UTF-8")
        self.__foreign_cryption = Fernet(self.__foreign_key)

    def new_key(self) -> str:
        self.__own_key = Fernet.generate_key()
        self.__own_cryption = Fernet(self.__own_key)
        return self.__own_key.decode("UTF-8")




"""
deamons/connection/encryption/_cryption_method.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################


##################################################
#                     Code                       #
##################################################

class CryptionMethod:
    """
    Every cryptionmethod should inherit from this and overwrite all functions
    """
    def encrypt(self, message: bytes) -> bytes:
        """
        Encrypt messages
        :param message: Raw message
        :return: Encrypted message
        """
        ...

    def decrypt(self, message: bytes) -> bytes:
        """
        Decrypte messages
        :param message: Encrypted message
        :return: Raw message
        """

    def get_key(self) -> str:
        """
        Get the main/public key to share with the other site
        :return: Key
        """
        ...

    def set_key(self, key: str) -> None:
        """
        Set the main/public key
        """
        ...

    def new_key(self) -> str:
        """
        Generate a new key and return public key
        """
        ...



"""
fridex/connection/encryption/_test_cryption.py

Project: Fridrich-Connection
Created: 12.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from string import printable
import unittest

from ._cryption_method import CryptionMethod
from ._cryption import CryptionService, CRYPTION_METHODS


##################################################
#                     Code                       #
##################################################

class CryptionTest(unittest.TestCase):
    """
    Test encryption
    """
    _cryption: CRYPTION_METHODS

    _left: CryptionMethod
    _right: CryptionMethod

    _raw_left = printable.encode()
    _raw_right = _raw_left[::-1]

    _encrypted_left: bytes
    _encrypted_right: bytes

    _decrypted_left: bytes
    _decrypted_right: bytes

    def setUp(self) -> None:
        """
        Create new cryptions
        """
        self._left = CryptionService.new_cryption(self._cryption)
        self._right = CryptionService.new_cryption(self._cryption)

    def dual_side_exchange(self) -> None:
        """
        Basic exchange in each side
        """
        self._encrypted_left = self._left.encrypt(self._raw_left)
        self._decrypted_left = self._right.decrypt(self._encrypted_left)

        self._encrypted_right = self._right.encrypt(self._raw_right)
        self._decrypted_right = self._left.decrypt(self._encrypted_right)

        self.assertEqual(self._raw_left, self._decrypted_left)
        self.assertEqual(self._raw_right, self._decrypted_right)

        print(self._encrypted_left, "\n", self._encrypted_right)

    def test_no_keys(self) -> None:
        """
        Test cryption without any keys set
        """
        self.dual_side_exchange()
        self.assertEqual(self._raw_left, self._encrypted_left)
        self.assertEqual(self._raw_right, self._encrypted_right)

    def test_keys(self) -> None:
        """
        Test cryption with both keys generated
        """
        self._left.set_key(self._right.get_key())
        self._right.set_key(self._left.get_key())
        self.dual_side_exchange()

        self.assertNotEqual(self._raw_left, self._encrypted_left)
        self.assertNotEqual(self._raw_right, self._encrypted_right)
        self.assertNotEqual(self._left.get_key(), self._right.get_key())

    def tearDown(self) -> None:
        """
        Cleanup used cryption_methods
        """
        del self._right
        del self._left



"""
deamons/connection/encryption/_cryption.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import Literal

from ._private_public import PrivatePublicCryption
from ._cryption_method import CryptionMethod
from ._fernet import FernetCryption

##################################################
#                     Code                       #
##################################################

CRYPTION_METHODS = Literal["private_public", "fernet"]


class CryptionService:
    """
    Encrypt and decrypt messages with different encryptions
    """

    @staticmethod
    def new_cryption(method: CRYPTION_METHODS = "private_public") -> CryptionMethod:
        match method:
            case "private_public":
                return PrivatePublicCryption()
            case "fernet":
                return FernetCryption()



"""
deamons/connection/protocol/_protocol.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from datetime import datetime
from typing import get_args
from json import dumps

from ._types import BulkDict, MessageDict, KINDS, DIRECTIONS, DATAUNIT


##################################################
#                     Code                       #
##################################################


class MessageToLongError(Exception):
    ...


class Protocol:
    """
    Default protocol for all kinds and directions of messages
    """
    max_bytes: int = 4
    max_size: int = 4294967296

    __bulks: dict[DIRECTIONS, list[MessageDict]]
    __kind: KINDS

    _id_range: range
    _id_count: int

    def __init__(
            self,
            kind: KINDS,
            id_range: range = range(100, 999)
    ) -> None:
        """
        Configure protocol and id system
        :param kind: Specify kind of this protocol
        :param id_range: Numrange for message id (0 - 1000 is reserved)
        """
        self.__kind = kind

        self._id_range = id_range
        self._id_count = id_range.start

        self.__bulks = {}
        for direction in get_args(DIRECTIONS):
            self.__bulks[direction] = []

    @staticmethod
    def set_max_bytes(value: int) -> None:
        """
        Set the bytes to communicate the maximum length for all protocols
        :param value: New value to set
        """
        Protocol.max_bytes = value
        Protocol.max_size = 2 ** (value * 8)

    @property
    def id_count(self) -> int:
        """
        :return: Current id count
        """
        return self._id_count

    def _encapsulate(
            self,
            message: DATAUNIT | list[MessageDict],
            direction: DIRECTIONS,
            single: bool = True,
            id_: int | None = None
    ) -> None | str:
        """
        Encapsulate message
        :param message: The message itself
        :param direction: Specify the message direction
        :param single: Whether it's a single message that should be added to the queue or the whole queue request
        :param id_: When no new ID should be used (when direction is response)
        :return: Depends on single
        :raises MessageToLongError: If message is too long to communicate length
        """
        # Base encapsulation
        additional_information: MessageDict | BulkDict = {
            "time": datetime.now().timestamp(),
            "data": message
        }

        # Additional encapsulation
        if single:
            additional_information: MessageDict
            additional_information["id"] = id_ if id_ is not None else self._id_count
        else:
            additional_information: BulkDict
            additional_information["direction"] = direction
            additional_information["kind"] = self.__kind

        # Increase id
        if not id_ and single:
            self._id_count += 1
            if self._id_count >= self._id_range.stop:
                self._id_count = self._id_range.start

        # Return | Save in bulk
        if single:
            self.__bulks[direction].append(additional_information)
        else:
            message_str: str = dumps(additional_information)

            if len(message_str) > self.__class__.max_size:
                raise MessageToLongError(f"With a size of {len(message_str)} the message is to long!")

            return message_str
        return None

    def request_start(self) -> None:
        """
        Start a bulk request message queue
        """
        self.__bulks["request"] = []

    def request_add(self, message: DATAUNIT) -> None:
        """
        Add a message to the request bulk queue
        :param message: Message to add
        """
        self._encapsulate(message, direction="request")

    def request_get(self, restart: bool = True) -> str | None:
        """
        Get the whole request message queue
        :param restart: Whether the request queue should be reseted
        :return: String to send
        """
        try:
            if self.__bulks["request"]:
                return self._encapsulate(self.__bulks["request"], direction="request", single=False)
            return None
        finally:
            if restart:
                self.request_start()

    def response_start(self) -> None:
        """
        Start a bulk response message queue
        """
        self.__bulks["response"] = []

    def response_add(self, message: DATAUNIT, id_: int) -> None:
        """
        Add a message to the response bulk queue
        :param message: Message to add
        :param id_: ID the response is marked with
        """
        self._encapsulate(message, direction="response", id_=id_)

    def response_get(self, restart: bool = True) -> str | None:
        """
        Get the whole response message queue
        :param restart: Whether the response queue should be reseted
        :return: String to send
        """
        try:
            if self.__bulks["response"]:
                return self._encapsulate(self.__bulks["response"], direction="response", single=False)
            return None
        finally:
            if restart:
                self.response_start()

    def process_response(self, message: BulkDict) -> None:
        """
        Every protocol should overwrite this function to process incomming responses
        """
        ...

    def process_request(self, message: BulkDict) -> str | None:
        """
        Every protocol should overwrite this function to process incomming requests
        """
        ...



"""
deamons/connection/protocol/_control.py

Project: Fridrich-Connection
Created: 30.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import TypedDict, Any, Callable, Literal

from ._types import BulkDict, MessageDict
from ._protocol import Protocol

from ..encryption import CRYPTION_METHODS


##################################################
#                     Code                       #
##################################################

class CommunicationData(TypedDict):
    type: Literal["key", "max_bytes", "cryption", "state"]
    value: Any


class CommunicationProtocol:
    """
    Protocol for connection changes
    All Messages are sent in "MasterSlave"-Mode
    """
    __protocol: Protocol

    NEW_KEY_CALLBACK_TYPE = Callable[[], str]
    SET_KEY_CALLBACK_TYPE = Callable[[str], Any]
    CONTROL_CALLBACK = Callable[[], Any]
    MAX_BYTES_CALLBACK_TYPE = Callable[[int], Any]
    CRYPTION_REQ_CALLBACK_TYPE = Callable[[str, str], tuple[NEW_KEY_CALLBACK_TYPE, SET_KEY_CALLBACK_TYPE]]
    CRYPTION_RES_CALLBACK_TYPE = Callable[[str], tuple[NEW_KEY_CALLBACK_TYPE, SET_KEY_CALLBACK_TYPE]]
    PAUSE_CONNECTION_CALLBACK_TYPE = Callable[[], Any]
    RESUME_CONNECTION_CALLBACK_TYPE = Callable[[], Any]

    __new_key_callback: NEW_KEY_CALLBACK_TYPE
    __set_key_callback: SET_KEY_CALLBACK_TYPE
    __control_callback: CONTROL_CALLBACK
    __max_bytes_callback: MAX_BYTES_CALLBACK_TYPE
    __cryption_req_callback: CRYPTION_REQ_CALLBACK_TYPE
    __cryption_res_callback: CRYPTION_RES_CALLBACK_TYPE
    __pause_connection_callback: PAUSE_CONNECTION_CALLBACK_TYPE
    __resume_connection_callback: RESUME_CONNECTION_CALLBACK_TYPE

    def __init__(
            self,
            id_range: range,
            new_key_callback: NEW_KEY_CALLBACK_TYPE,
            set_key_callback: SET_KEY_CALLBACK_TYPE,
            control_callback: CONTROL_CALLBACK,
            max_bytes_callback: MAX_BYTES_CALLBACK_TYPE,
            cryption_req_callback: CRYPTION_REQ_CALLBACK_TYPE,
            cryption_res_callback: CRYPTION_RES_CALLBACK_TYPE,
            pause_connection_callback: PAUSE_CONNECTION_CALLBACK_TYPE,
            resume_connection_callback: RESUME_CONNECTION_CALLBACK_TYPE
    ) -> None:
        """
        Create communication protocol
        :param id_range: ID range to use for this subprotocol
        :param new_key_callback: Callback to generate a new key and get it
        :param set_key_callback: Callback when a new key is received
        :param control_callback: Callback when control response is received that requires pausing the connection
        :param max_bytes_callback: Callback to set new max bytes
        :param cryption_req_callback: Callback to set new encryption with key.
        :param cryption_res_callback: Callback to confirm cryption change and set key
        :param pause_connection_callback: Callback to pause connection if requested
        :param resume_connection_callback: Callback to resume connection if requested
        """
        self.__protocol = Protocol("com", id_range)
        self.__new_key_callback = new_key_callback
        self.__set_key_callback = set_key_callback
        self.__control_callback = control_callback
        self.__max_bytes_callback = max_bytes_callback
        self.__cryption_req_callback = cryption_req_callback
        self.__cryption_res_callback = cryption_res_callback
        self.__pause_connection_callback = pause_connection_callback
        self.__resume_connection_callback = resume_connection_callback

    def __response(self, data: CommunicationData, id_: int) -> str:
        """
        General response encapsulation
        :param data: Response dictonary
        :param id_: ID of the conversation
        :return: Raw string to send
        """
        self.__protocol.response_add(data, id_)
        return self.__protocol.response_get()

    def __request(self, data: CommunicationData) -> str:
        """
        General request encapsulation
        :param data: Dictonary to encapsulate
        :return: Raw string to send
        """
        self.__protocol.request_add(data)
        return self.__protocol.request_get()

    def request_key_exchange(self) -> str:
        """
        Request exchanging encryption keys
        :return: Raw string to send
        """
        return self.__request({"type": "key", "value": self.__new_key_callback()})

    def _response_key_exchange(self, id_: int) -> str:
        """
        Also send back a new key
        :param id_: ID of the conversation
        :return: Confirm message
        """
        return self.__response({"type": "key", "value": self.__new_key_callback()}, id_)

    def request_max_bytes(self, num: int) -> str:
        """
        Redefine number of bytes to communicate length
        :param num: Number of bytes
        :return: String to send
        """
        self.__protocol.set_max_bytes(num)
        return self.__request({"type": "max_bytes", "value": num})

    def request_crpytion(self, new_cryption: CRYPTION_METHODS, new_key: str | None = None) -> str:
        """
        Request to change the encryption
        :param new_cryption: New encryption to use
        :param new_key: Key for the new encryption
        :return: String to send
        """
        return self.__request({"type": "cryption", "value": {"name": new_cryption, "key": new_key}})

    def _response_cryption(self, id_: int) -> str:
        """
        Response to cryption change with a key
        :param id_: ID of the conversation
        :return: String to send
        """
        return self.__response({"type": "cryption", "value": self.__new_key_callback()}, id_)

    def request_pause(self) -> str:
        """
        Request to pause connection
        :return: String to send
        """
        return self.__request({"type": "state", "value": "pause"})

    def _response_pause(self, id_: int) -> str:
        """
        Respone to pause connection request
        :param id_: ID of the conversation
        :return: String to send
        """
        return self.__response({"type": "state", "value": "pause"}, id_)

    def request_resume(self) -> str:
        """
        Request to resume connection
        :return: String to send
        """
        return self.__request({"type": "state", "value": "resume"})

    def _response_resume(self, id_: int) -> str:
        """
        Respone to resume connection request
        :param id_: ID of the conversation
        :return: String to send
        """
        return self.__response({"type": "state", "value": "resume"}, id_)

    def process_response(self, message: BulkDict) -> None:
        """
        Process incoming communication responses
        :param message: Response message
        """
        submessage: MessageDict = message["data"][0]
        match submessage["data"]["type"]:
            case "key":
                self.__set_key_callback(submessage["data"]["value"])
                self.__control_callback()
            case "cryption":
                self.__new_key_callback, self.__set_key_callback = \
                    self.__cryption_res_callback(submessage["data"]["value"])
                self.__control_callback()
            case "state":
                self.__control_callback()

    def process_request(self, message: BulkDict) -> str | None:
        """
        Process incoming communication requests
        :param message: Request message
        :return: Response message
        """
        submessage: MessageDict = message["data"][0]
        id_: int = submessage["id"]

        match submessage["data"]["type"]:
            case "key":
                self.__set_key_callback(submessage["data"]["value"])
                return self._response_key_exchange(id_)

            case "max_bytes":
                self.__max_bytes_callback(submessage["data"]["value"])

            case "cryption":
                self.__new_key_callback, self.__set_key_callback = \
                    self.__cryption_req_callback(submessage["data"]["value"]["name"],
                                                 submessage["data"]["value"]["key"])
                return self._response_cryption(id_)

            case "state":
                match submessage["data"]["value"]:
                    case "pause":
                        self.__pause_connection_callback()
                        return self._response_pause(id_)
                    case "resume":
                        self.__resume_connection_callback()
                        return self._response_resume(id_)

        return None



"""
fridex/connection/protocol/_connection.py

Project: Fridrich-Connection
Created: 13.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import TypedDict, Literal, Callable, Any

from ._types import BulkDict, MessageDict
from ._protocol import Protocol


##################################################
#                     Code                       #
##################################################

class ControlData(TypedDict):
    type: Literal["ping", "alive"]


class ControlProtocol:
    """
    Protocol for connection control
    """
    __protocol: Protocol

    PING_CALLBACK_TYPE = Callable[[], Any]

    __ping_callback: PING_CALLBACK_TYPE

    def __init__(
            self,
            id_range: range,
            ping_callback: PING_CALLBACK_TYPE
    ) -> None:
        """
        Create control protocol
        :param id_range: ID range to use for this subprotocol
        :param ping_callback: Callback when ping response is received
        """
        self.__protocol = Protocol("con", id_range)
        self.__ping_callback = ping_callback

    def __response(self, data: ControlData, id_: int) -> str:
        """
        General response encapsulation
        :param data: Response dictonary
        :param id_: ID of the conversation
        :return: Raw string to send
        """
        self.__protocol.response_add(data, id_)
        return self.__protocol.response_get()

    def __request(self, data: ControlData) -> str:
        """
        General request encapsulation
        :param data: Dictonary to encapsulate
        :return: Raw string to send
        """
        self.__protocol.request_add(data)
        return self.__protocol.request_get()

    def request_ping(self) -> str:
        """
        Request ping eachother
        :return: Ping string to send
        """
        return self.__request({"type": "ping"})

    def _response_ping(self, id_: int) -> str:
        """
        Confirm ping request
        :param id_: ID of the conversation
        :return: Confirm message
        """
        return self.__response({"type": "ping"}, id_)

    def request_alive(self) -> str:
        """
        Provide alive message
        :return: String to send
        """
        return self.__request({"type": "alive"})

    def process_response(self, message: BulkDict) -> None:
        """
        Process incoming control responses
        :param message: Response message
        """
        submessage: MessageDict = message["data"][0]
        match submessage["data"]["type"]:
            case "ping":
                self.__ping_callback()

    def process_request(self, message: BulkDict) -> str | None:
        """
        Process incoming control requests
        :param message: Request message
        :return: Response message
        """
        submessage: MessageDict = message["data"][0]
        id_: int = submessage["id"]

        match submessage["data"]["type"]:
            case "ping":
                return self._response_ping(id_)
            case "alive":
                self.__ping_callback()



"""
deamons/connection/communication/_server_connection.py

Project: Fridrich-Connection
Created: 26.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

import socket

from ..protocol import ProtocolInterface, DATAUNIT
from ._base_connection import BaseConnection


##################################################
#                     Code                       #
##################################################

class ServerConnection(BaseConnection):
    """
    Connection from the server to the client
    """
    def __init__(
            self,
            conn: socket.socket,
            request_callback: ProtocolInterface.REQUEST_CALLBACK_TYPE,
            rework_callback: ProtocolInterface.REWORK_CALLBACK_TYPE,
            add_sub_callback: ProtocolInterface.ADD_RELATED_SUB_CALLBACK_TYPE,
            del_sub_callback: ProtocolInterface.DELETE_RELATED_SUB_CALLBACK_TYPE,
    ) -> None:
        """
        Create connection
        :param conn: Socket
        :param request_callback: Callback to get information for data requests
        :param rework_callback: Callback to rework result before setting to future
        :param add_sub_callback: Callback when an add subscription request comes in
        :param del_sub_callback: Callback when a delete subscription request comes in
        """
        super().__init__(conn,
                         request_callback=request_callback,
                         rework_callback=rework_callback,
                         add_sub_callback=add_sub_callback,
                         del_sub_callback=del_sub_callback)

    def provide_data(self, req_dict: DATAUNIT, value: DATAUNIT) -> None:
        """
        Check if data is needed in a subscription
        :param req_dict: Same dictonary as a normal request to use
        :param value: New value
        """
        self._protocol.subscription.provide_data(req_dict, value)



"""
deamons/connection/protocol/_data.py

Project: Fridrich-Connection
Created: 31.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from concurrent.futures import Future
from json import loads
from typing import Callable

from ._types import BulkDict, DATAUNIT
from ._protocol import Protocol
from ._cache import Cache


##################################################
#                     Code                       #
##################################################

class DataProtocol(Protocol):
    """
    Protocol for custom traffic
    """
    __cache: Cache | None

    REQUEST_CALLBACK_TYPE = Callable[[DATAUNIT], DATAUNIT]
    REWORK_CALLBACK_TYPE = Callable[[DATAUNIT], DATAUNIT]

    __request_callback: REQUEST_CALLBACK_TYPE
    __rework_callback: REWORK_CALLBACK_TYPE

    __send_futures: dict[int, Future]

    def __init__(
            self,
            id_range: range,
            request_callback: REQUEST_CALLBACK_TYPE,
            rework_callback: REWORK_CALLBACK_TYPE,
            cache: Cache | None = None
    ) -> None:
        """
        Create data protocol
        :param id_range: ID range to use for this subprotocol
        :param request_callback: Callback to get information for requests
        :param rework_callback: Callback to rework result before setting to future
        :param cache: Optional cache
        """
        super().__init__("data", id_range)
        self.__request_callback = request_callback
        self.__rework_callback = rework_callback
        self.__cache = cache

        self.__send_futures = {}

    def request_add(self, message: DATAUNIT) -> Future:
        """
        Add a message to the request bulk queue
        :param message: Message to add
        :return: Future instance to receive the result
        """
        future: Future = Future()

        if self.__cache:
            value = self.__cache.get(message)
            if value:
                future.set_result(self.__rework_callback(value))
                return future

        super().request_add(message)

        self.__send_futures[self._id_count-1] = future
        return future

    def process_response(self, message: BulkDict) -> None:
        """
        Supply Futures results with response data
        :param message: Response message
        """
        for submessage in message["data"]:
            value: DATAUNIT = self.__rework_callback(loads(submessage["data"]))
            self.__send_futures[submessage["id"]].set_result(value)
            self.__send_futures.pop(submessage["id"])

    def process_request(self, message: BulkDict) -> str:
        """
        Process data request and get all required information
        :param message: Request message
        :return: Response message with all information
        """
        self.response_start()

        for sub_req in message["data"]:
            self.response_add(self.__request_callback(sub_req["data"]), id_=sub_req["id"])

        return self.response_get()



"""
deamons/connection/communication/_client_connection.py

Project: Fridrich-Connection
Created: 26.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import Callable, Any
import socket

from ._base_connection import BaseConnection
from ..protocol import ProtocolInterface


##################################################
#                     Code                       #
##################################################

class ClientConnection(BaseConnection):
    """
    Connection from the client to the server
    """
    def __init__(
            self,
            ip: str,
            port: int,
            request_callback: ProtocolInterface.REQUEST_CALLBACK_TYPE,
            rework_callback: ProtocolInterface.REWORK_CALLBACK_TYPE
    ) -> None:
        """
        Connect to server
        :param ip: IP of the server
        :param port: Port to connect
        :param request_callback: Callback to get information for data requests
        :param rework_callback: Callback to rework result before setting to future
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((ip, port))

        super().__init__(conn=sock, request_callback=request_callback, rework_callback=rework_callback)

        self._state = "open"
        self.send_key_exchange()
        self.send_key_exchange()

    def add_subscription(
            self,
            callback: Callable[[Any], Any],
            request_dict: dict[str | int | float | bool | None, any]
    ) -> int:
        """
        Send add subscription request
        :param callback: Callback when value is updated
        :param request_dict: Same dictonary as a normal request to use
        :return: ID of the subscripton
        """
        sub_id, message = self._protocol.subscription.add_subscription(callback, request_dict)
        self.send(message)

        return sub_id

    def delete_subscription(self, sub_id: int) -> None:
        """
        Send subscription delete request
        :param sub_id: ID of the subscription
        """
        self.send(self._protocol.subscription.remove_subscription(sub_id))



"""
deamons/connection/protocol/_protocol_types.py

Project: Fridrich-Connection
Created: 30.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from typing import TypedDict, Literal


##################################################
#                     Code                       #
##################################################

# Defines the different protocols that can be sent
KINDS = Literal["data", "sub", "con", "com"]
DIRECTIONS = Literal["request", "response"]
DATAUNIT = dict[str | int | float | bool | None, any]


class _Dict(TypedDict):
    """
    Shared information of MessageDict and BulkDict
    """
    time: float


class MessageDict(_Dict):
    """
    A MessageDict is always sent in a BulkDict
    """
    id: int
    data: DATAUNIT


class BulkDict(_Dict):
    """
    This is the dict that is really being sent
    """
    data: list[MessageDict]
    kind: KINDS
    direction: DIRECTIONS



"""
deamons/connection/communication/_base_connection_test.py

Project: Fridrich-Connection
Created: 01.06.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################


from concurrent.futures import Future
from time import sleep, time
from typing import Literal
from json import dumps
import unittest
import socket

from ._base_connection import BaseConnection, ProtocolInterface
from ..protocol import MessageToLongError


##################################################
#                     Code                       #
##################################################

server: socket.socket | None = None


def get_socket_pair(
        port: int = 12345
) -> tuple[socket.socket, socket.socket, socket.socket]:
    """
    Get a connected socket pair
    :param port: Port to use
    :return: Server, ServerConnection, Client
    """
    global server
    if server is None:
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.bind(("0.0.0.0", 12345))
        server.listen()

    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(("127.0.0.1", port))

    sock, add = server.accept()
    return server, sock, client


class BaseConnectionModified(BaseConnection):
    def set_state(self, state: Literal["init", "open", "closed"]) -> None:
        self._set_state(state)

    @property
    def protocol(self) -> ProtocolInterface:
        return self._protocol


class BaseConnectionTest(unittest.TestCase):
    """
    Test BaseConnection
    """
    server: socket.socket
    client_handler: socket.socket
    client: socket.socket

    conn_server: BaseConnectionModified
    conn_client: BaseConnectionModified

    def setUp(self) -> None:
        """
        Setup connections and BaseConnections
        """
        self.server, self.client_handler, self.client = get_socket_pair()

        self.conn_server = BaseConnectionModified(self.client_handler, lambda a: a, lambda a: a, timeout=1)
        self.conn_client = BaseConnectionModified(self.client, lambda a: a, lambda a: a, timeout=1)
        self.conn_server.set_state("open")
        self.conn_client.set_state("open")

    def hold_connection(self, duration: int | None = 5) -> None:
        """
        Hold the connection and then check if it's still open
        :param duration: Duration to hold in seconds
        """
        start = time()
        while time() < start + duration:
            sleep(0.1)

        self.assertEqual(self.conn_client.state, "open")
        self.assertEqual(self.conn_server.state, "open")

    def test_unencrypted_data(self) -> None:
        """
        Test basic unencrypted data traffic
        """
        f_client: list[tuple[Future, str]] = []
        f_server: list[tuple[Future, str]] = []

        for i in range(10):  # noqa
            self.conn_client.protocol.data.request_start()
            f_client.append((self.conn_client.protocol.data.request_add(dumps({"test": str(i)*1000})), str(i)*1000))
            self.conn_client.send(self.conn_client.protocol.data.request_get())

        for i in range(10):  # noqa
            self.conn_server.protocol.data.request_start()
            f_server.append((self.conn_server.protocol.data.request_add(dumps({"test": str(i)*1000})), str(i)*1000))
            self.conn_server.send(self.conn_server.protocol.data.request_get())

        for f, num in f_client + f_server:
            self.assertEqual(f.result()["test"], num)

        self.hold_connection(0)

    def test_encrypted_data(self) -> None:
        """
        Test encrypted data protocol traffic
        """
        self.conn_client.send_key_exchange()
        self.conn_client.send_key_exchange()

        self.test_unencrypted_data()

        self.hold_connection()

    def test_maxbytes(self) -> None:
        """
        Test max_bytes change
        """
        for num in [10, 0]:
            if num == 10:
                req = self.conn_client.protocol.communication.request_max_bytes(num)
            else:
                self.assertRaises(MessageToLongError, lambda: self.conn_client.protocol.communication.request_max_bytes(num))
                break

            self.conn_client.send(req)

            self.hold_connection()

            self.assertEqual(self.conn_client.protocol.data.max_bytes, num)
            self.assertEqual(self.conn_server.protocol.data.max_bytes, num)

            self.hold_connection(2)

    def test_cryption(self) -> None:
        """
        Test cryption change
        """
        self.conn_client.send_cryption_change("fernet")

        self.hold_connection()

    def tearDown(self) -> None:
        """
        Clearup after each test
        """
        self.conn_client.close()
        self.conn_server.close()

        del self.client, self.client_handler, self.conn_server, self.conn_client



"""
deamons/connection/encryption/__init__.py

Project: Fridrich-Connection
Created: 25.05.2023
Author: Lukas Krahbichler
"""

from ._cryption import CryptionService, CRYPTION_METHODS
from ._private_public import PrivatePublicCryption
from ._cryption_method import CryptionMethod
from ._fernet import FernetCryption



"""
/setup.py

Project: Fridrich-Connection
Created: 23.05.2023
Author: Lukas Krahbichler
"""

##################################################
#                    Imports                     #
##################################################

from setuptools import find_packages, setup
from pathlib import Path


##################################################
#                     Code                       #
##################################################

here = Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="fridex-connection",
    version="0.0.1",
    description="Basic fridrich functions (connection, encryption , ...)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Lukas Krahbichler",
    author_email="melonenbuby@proton.me",
    url="https://github.com/FridrichDerGosse/Fridrich-Connection",
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keyword="dashboard, voting, connection",
    packages=find_packages(),
    namespace_packages=['fridex'],
    python_requires=">=3.11"
)
